/*************************************************/
/* Author: Kyri Pavlou           15 January 2009 */
/*************************************************/

#include "algorithms.h"
int *C_set_RGBY;   		        /*the candidate set*/
int c_RGBY_capacity = INIT_CAP; 	/*the capacity of the candidate set*/
int c_RGBY_logical = ZERO; 	        /*the logical size of the candidate set*/

/***************************************************/
/* Input                                           */
/*  t_FVF: the time of first validation failure    */
/*  I_N: notarization interval                     */
/*  R_s: spatial detection resolution              */
/* Output                                          */
/*  C_set_RGBY the candidate set                   */
/*  USB upper  spatial bound                       */
/*  LSB lower  spatial bound                       */
/*  UTB upper  temporal bound                      */
/*  LTB lower  temporal bound                      */
/***************************************************/
void RGBY(int t_FVF, int I_N, int R_s)
{
 	int  I_V, t_RVS, V, USB, LSB, UTB, LTB, u, n, s, i;
					/*chains used in the linear search of the partial chains*/
	Chain currChain, lastChain;	/*currChain is the chain currently under consideration*/
					/*lastChain is the last partial chain computed at t_FVF*/			
	Chain *succChain = (Chain *)mymalloc(sizeof(Chain));            /*the partial chain immediately after currChain, i.e., its successor*/
	V = 2; 								/*the validation factor is fixed for the RGBY algorithm*/
	C_set_RGBY = (int *)mymalloc(c_RGBY_capacity * sizeof(int)); 	/*the candidate set storing the corrupted granules*/
	I_V = V * I_N;
        t_RVS = find_t_RVS(t_FVF, I_N); 				/*find time of most recent validation success*/ 
	DBG(printf("t_RVS is %d\n", t_RVS));
	/*set the spatial and temporal bounds*/
	USB = t_RVS + I_N;
 	LSB = t_RVS;
    	UTB = t_FVF;
   	LTB = MAX(t_FVF - I_V, t_RVS);
	DBG(printf("USB = %d, LSB = %d, UTB = %d, LTB = %d\n", USB, LSB, UTB, LTB));
	append_RGBY(&C_set_RGBY, t_RVS+1);
	u = t_FVF/I_V;				/*the subscript of the lastChain*/
	lastChain.subscript = u;                
	lastChain.color = 1 + (u%2);            /*knowing the subscript we can compute the color of the lastChain*/
	DBG(printf("last chain is %c_%d\n", number_to_color(lastChain.color), lastChain.subscript));
	n = LSB / I_N;				
	s = ceil(n/2.0) + 1;		
	currChain.color = (n+3)% COLOR_NUM;     /*currChain starts as the most recent (i.e., rightmost)*/
	currChain.subscript = s;                /*partial chain covering the oldest corruption site*/
	DBG(printf("current chain is %c_%d\n", number_to_color(currChain.color), currChain.subscript));
	
	int val_succ = FALSE;                   /*boolean used for storing locally the validation result of the successor 
						  from the previous iteration of the "while" loop. This is an optimization 
						  because it guarantees that all partial chains during the linear search 
                                                  are val_checked at most once*/
	/*linearly scan all partial chains to look for multiple corruption sites*/
	/*iterate while the current chain comes before the last chain*/
	while(less_or_equal(&currChain, &lastChain)) 
	{					     /*the look-ahead is one so check current chain and its successor*/
		successor(&succChain, &currChain);   /*need to validate two chains at each iteration seeking the pattern ...False False...*/
		if(!val_succ || !val_check_RGBY(t_FVF, I_V, &currChain)) /*short-circuit by checking value of the successor from the previous iteration*/
		{	
			val_succ = val_check_RGBY(t_FVF, I_V, succChain); /*store locally the validation result of the successor*/
			if(!val_succ)
			{	
				if(currChain.color == BLUE || currChain.color == RED)    
					append_RGBY(&C_set_RGBY, 2*(currChain.subscript-1)*R_s + 1);
				else
					append_RGBY(&C_set_RGBY, 2*currChain.subscript*R_s - (R_s-1));
				
			}
		}
		currChain.color = succChain->color;
		currChain.subscript = succChain->subscript;
		DBG(printf("while loop after assignment current chain is %c_%d\n\n", number_to_color(currChain.color), currChain.subscript));			
	}

	/*print the corrupted granules; if R_s>1 then output intervals*/
	printf("\n C_set = { ");
	for(i = 0; i < c_RGBY_logical ; i++)
	{
		if(R_s > 1)
		{
			printf(" [%d,%d] ", C_set_RGBY[i], C_set_RGBY[i] + R_s - 1 );
		}
		else
		{
			printf(" %d ", C_set_RGBY[i]);
		}
	}
	printf(" }\n");
	free(C_set_RGBY);
	free(succChain);
}

/***************************************************************************/
/* Input                                                                   */
/*  chain1: a partial chain 						   */
/*  chain2: a partial chain                                 		   */
/* Output                                                                  */
/*  Returns the boolean result True if chain1 is the same as chain2        */
/*  or chain1 occurs chronologically before chain2.                        */
/*  Otherwise it returns false.                                            */
/*********************************************************************** ***/
int less_or_equal(Chain *chain1, Chain *chain2)
{
	int result = FALSE;
	if((chain1->subscript < chain2->subscript) || ((chain1->subscript == chain2->subscript) && (chain1->color < chain2->color)))
		result = TRUE;
	else if ((chain1->subscript == chain2->subscript) && (chain1->color == chain2->color))
		result = TRUE;
	return result;
}


/***************************************************/
/* Input                                           */
/*  t_FVF: time of first validation failure        */
/*  I_V:   validation interval                     */
/*  test:  partial hash chain to be tested         */
/* Output                                          */
/*  Returns the validation result (boolean) of     */
/*  the hash chain test.                           */  
/***************************************************/
int val_check_RGBY(int t_FVF, int I_V, Chain *test)
{
    counter_RGBY_f++;
    DBG(printf("Valcheck chain %c_%d \t", number_to_color(test->color), test->subscript));
    int result =  *(RGBYchains + (test->color)*(t_FVF/I_V) + test->subscript);
    DBG(printf("result %d\n", result));
	return result;
}


/***************************************************************************/
/* Input                                                                   */
/*  succChain: the successor, i.e., the chain immediately after currChain  */
/*  currChain: the current partial chain                                   */
/* Output                                                                  */
/*  Computes and sets the color and subscript of the successor             */
/*  of currChain, i.e., the partial chain immmediately after currChain.    */
/***************************************************************************/
void successor(Chain **succChain, Chain *currChain)
{
	DBG(printf("in successor function curr.chain is %c_%d\n", number_to_color(currChain->color), currChain->subscript));
	/*transitioning from green to red or from yellow to blue requires that the subscript be incremented*/
	if((currChain->color==GREEN) || ((currChain->color==YELLOW))) 
	{
		DBG(printf("current chain is green or yellow\n"));
		(*succChain)->subscript = (currChain->subscript)+1;
	}
	else
	{
		(*succChain)->subscript = currChain->subscript;
	}
	(*succChain)->color = (currChain->color + 1)%COLOR_NUM;
	DBG(printf("in successor function succ.chain is %c_%d\n", number_to_color((*succChain)->color), (*succChain)->subscript));
}


/***************************************************************/
/* Appends corrupted granule g to candidate set C_set_RGBY     */
/* If there is not enough space in the set the array is        */
/* resized to double its previous size.                        */
/***************************************************************/
void append_RGBY(int **C_set_RGBY, int g)
{
    DBG(printf("Appending %d to candidate set...\n", g));
    if(c_RGBY_logical < c_RGBY_capacity)
    {
        (*C_set_RGBY)[c_RGBY_logical] = g;
    }
    else
    {
        DBG(printf("Resizing the candidate set...\n"));
        int i;
        int *C_temp;
        c_RGBY_capacity *= 2;
        C_temp = (int *)mymalloc(c_RGBY_capacity*sizeof(int));
        for(i = 0; i < c_RGBY_logical; i++)
        {
            C_temp[i] = (*C_set_RGBY)[i];
        }
        C_temp[c_RGBY_logical] = g;
        free(*C_set_RGBY);
        *C_set_RGBY = C_temp;
    }
    c_RGBY_logical++;
}

/****************************************************************/
/* Input                                                        */
/*  num: the integer which encodes the color of a partial chain */
/* Output                                                       */
/*  Returns the color char corresponding to num                 */
/*  (B)lue = 0, (G)reen = 1, (R)ed = 2, (Y)ellow = 3            */
/****************************************************************/
char number_to_color(int num)
{
	char color;
	switch (num)
	{
		case 0:
		{	
			color = 'B';
			break;
		}
		case 1:
		{	
			color = 'G';
			break;
		}
		case 2:
		{	
			color = 'R';
			break;
		}
		case 3:
		{	
			color = 'Y';
			break;
		}
	}	
	return color;
}



