/*************************************************/
/* Author: Kyri Pavlou           16 January 2009 */
/*************************************************/

#include "algorithms.h"

int *t_l;                   /*the int array with all the time loci of the granules to be corrupted*/

int main(int argc, char **argv)
{
    char alg_flag[FLAG_SIZE];
    int i,j,k;
    int ce_num = 0;         /*the number of locus times t_l; also the true number of corruptions*/
    if(argc < MIN_NUM_ARGS) 
    {
        fprintf(stderr, "\nUsage: forensic <R_s> <algorithm> <N> <V> <t_FVF> <t_l1> [<t_l2> ...]\n");
        fprintf(stderr, "<algorithm>: \n\t type m for monochromatic\n\t type r for RGBY\n\t type t for tiled-bitmap\n\t type a for a3D\n");
        exit(1);
    }
    R_s = atoi(argv[RT_INDEX]);
    for(i = 0; i < FLAG_SIZE;i++)
	alg_flag[i] = '\0';
    strncpy(alg_flag, argv[ALG_INDEX],FLAG_SIZE);
    if(strncmp(alg_flag, "m",FLAG_SIZE) == ZERO)
    {
        N = 1;
        V = atoi(argv[V_INDEX]);
    }
    else if (strncmp(alg_flag, "r", FLAG_SIZE) == ZERO)
    {
        N = atoi(argv[N_INDEX]);
        V = atoi(argv[V_INDEX]);
	if(N != 1 || V != 2)
	{
	    DBG(printf("\n\nFor the RGBY algorithm V=2 and N=1 always\n"));
	    N = 1;		
	    V = 2;
	}
    }
    else if (strncmp(alg_flag, "t", FLAG_SIZE) == ZERO)
    {
        N = atoi(argv[N_INDEX]);
        V = atoi(argv[V_INDEX]);
        I_N = N * R_s;
        num_chains_per_tile = 1+log_base_n(I_N/R_s,2);
        DBG(printf("num_chains_per_tile = %d\n", num_chains_per_tile));
    }
    else if(strncmp(alg_flag, "a", FLAG_SIZE) == ZERO)
    {
        N = atoi(argv[N_INDEX]);
        V = 1;
    }
    else 
    {
        fprintf(stderr,"Invalid algorithm\n");
        exit(1);
    }
    t_FVF = atoi(argv[TFVF_INDEX]);
    /*calculate I_V and I_N*/
    I_N = N * R_s;
    I_V = V * I_N;
    /*Verify that no command line input is zero*/
    if(R_s == ZERO || V == ZERO || N == ZERO || I_N == ZERO || I_V == ZERO || t_FVF == ZERO)
    {
        fprintf(stderr, "Zero-valued parameter. Cannot proceed...\n");
        exit(1);
    }
    /*Verify that t_FVF is a multiple of I_V*/
    if(t_FVF % I_V != ZERO)
    {
        fprintf(stderr, "t_FVF must be an integer multiple of I_V\n");
        exit(1);
    }
    /*Verify that N is a power of 2*/
    if((N & (N - 1)) != 0)
    {
        fprintf(stderr, "the notarization factor N must be a power of 2\n");
        exit(1);
    }

    /*Verify that all 0 <= t_l's <= t_FVF*/
    for(i = TL_START; i <  argc; i++)
    {
        if(atoi(argv[i]) > t_FVF)
        {
            fprintf(stderr, "All t_l's must be <= to t_FVF = %d\n", t_FVF);
            exit(1);
        }
        if(atoi(argv[i]) <= 0)
        {
            fprintf(stderr, "All t_l's must be positive\n");
            exit(1);
        }
    }
    /*initialize the normal porcessing and forensic analysis counters*/
    counter_mono_n = counter_RGBY_n = counter_tiled_n = counter_a3D_n = 0;
    counter_mono_f = counter_RGBY_f = counter_tiled_f = counter_a3D_f = 0;

    /*Inform user of the configuration*/
    DBG(printf("\nConfiguration: Algorithm = %s\n R_s = %d, V = %d, N = %d, I_N = %d, I_V = %d\n", alg_flag, R_s, V, N, I_N, I_V));
    DBG(printf("\nFirst Validation Failure = %d\n", t_FVF));

    //Parse all the time loci t_l and put them in the t_l array
    ce_num = argc-(MIN_NUM_ARGS -1);
    DBG(printf("t_l = {"));
    t_l = (int *)mymalloc((ce_num)*sizeof(int));
    for(i = TL_START; i <  argc; i++)
    {
        t_l[i-TL_START] = atoi(argv[i]);
        DBG(printf(" %d ", t_l[i-TL_START]));
    }
    DBG(printf("}\n"));

    BlackChains = (int *)mymalloc((t_FVF/I_N)*sizeof(int));
    DBG(printf("\nThere are %d black chains\n", t_FVF/I_N ));

    /*Initialize and tamper appropriate data structures depending on the selected forensic algorithm and provided t_ls*/
    /*Initialize the Black hash chains data structure*/
    DBG(printf("Initializing the black chains data structure \n\nBlackChains = {"));
    for (i = 0; i < t_FVF/I_N; i++)
    {
        BlackChains[i] = 1;
        DBG(printf(" %d ", BlackChains[i]));
	if((i+1) % V == 0) /*update counters for validations*/
	{	
		counter_tiled_n++;
		counter_a3D_n++;
		counter_RGBY_n++;
		counter_mono_n++;
	}
	/*update counters for notarizations*/
	counter_mono_n++;
	counter_RGBY_n++;
	counter_tiled_n++;
	counter_a3D_n++;
    }
    DBG(printf("}\n\n"));

    /*Tampering of black chains*/
    int min_tl = t_l[0]; /*min_tl will hold the CE with the oldest locus time*/
    for(i = 0; i < ce_num; i++)
    {
        if(t_l[i] < min_tl)
        {
            min_tl = t_l[i];
        }
    }
    DBG(printf("The oldest locus time is %d\n", min_tl));
    for ( i = ceil(((float)min_tl)/I_N) - 1 ; i < t_FVF/I_N ;  i++)
    {
        BlackChains[i] = 0; /*put in array the appropriate validation results 
			    (of the back chains) consistent with the CE locus times*/
    }
    /*Print the black chains validation results*/
    DBG(printf("The Black chains after tampering BlackChains = {"));
    for ( i = 0; i < t_FVF/I_N; i++)
    {
        DBG(printf(" %d ",BlackChains[i]));
    }
    DBG(printf("}\n"));

    /*Initialize and tamper the Red, Green, Blue, Yellow partial hash chains data structure for the RGBY algorithm*/
    if(strncmp(alg_flag, "r", FLAG_SIZE) == ZERO)
    {
	/*The correspondence between the colors of the chains and the RGBYchains first index is as follows Blue = 0, Green = 1, Red = 2, Yellow = 3*/
	RGBYchains = (int *)mymalloc((COLOR_NUM*(t_FVF/I_V)+1)*sizeof(int));   /*array with validation results of the partial RGBY chains*/
        DBG(printf("Initializing the RGBY chains data structure...\n"));
 	for( i = 0 ; i < COLOR_NUM ; i++)
        {
            for( j = 1; j <= (t_FVF/I_V) ; j++)
            {
		if(((i == RED || i == YELLOW) && (j%2 == 1)) || ((i == BLUE || i == GREEN) && (j%2 == 0)))
	        {
		    *(RGBYchains + (i*(t_FVF/I_V)) + j) = 1;  /*initialization of the RGBYchains data structure*/  
		    counter_RGBY_n++;
		}
            }
        }

	/*Tamper RGBYchains data structure. The validation results 
	should be consistent with the provided CE locus times t_ls*/
	int n, s;
	for(i = 0; i < ce_num; i++)
        {
	    n = ceil(((float)t_l[i])/I_N);
	    s = (n+1)/2;
	    if(n%2==1) /*both chains have the same subscript*/
	    {
	        if(s%2==1) /*Red and Yellow chains*/
		{
		    *(RGBYchains + RED*(t_FVF/I_V) + s) = 0;
		    *(RGBYchains + YELLOW*(t_FVF/I_V) + s) = 0;
		}
		else if(s%2==0) /*Blue and Green chains*/
		{
		    *(RGBYchains + BLUE*(t_FVF/I_V) + s) = 0;
		    *(RGBYchains + GREEN*(t_FVF/I_V) + s) = 0;
		}
		else
		{
		    fprintf(stderr, "Error...Abort...\n");
		    exit(1);
		}

	    }
	    else  /*n%2==0, i.e., the two chains have different subscripts*/
	    {
	        if(s%2==1) /*Yellow and Blue chains*/
		{
		    *(RGBYchains + YELLOW*(t_FVF/I_V) + s) = 0;
		    *(RGBYchains + BLUE*(t_FVF/I_V) + s + 1) = 0; 
		}
		else if(s%2==0) /*Green and Red chains*/
		{
		    *(RGBYchains + GREEN*(t_FVF/I_V) + s) = 0;
		    *(RGBYchains + RED*(t_FVF/I_V) + s + 1) = 0;
		}
		else
		{
		    fprintf(stderr, "Error...Abort...\n");
		    exit(1);
		}
	    }
	}

	/*Print the RGBY chains data structure after tampering has occurred*/
        DBG(printf("The RGBY data structure after tampering \n"));
        for( j = 1; j <= (t_FVF/I_V) ; j++ )
        {
            for( i = 0 ; i < COLOR_NUM ; i++ )
            {               
		if((i == RED) && (j%2 == 1))
		{	
		    DBG(printf("Red_%d = %d\n", j, *(RGBYchains + i*(t_FVF/I_V) + j)));
		}
		else if((i == YELLOW) && (j%2 == 1))
		{
		    DBG(printf("Yellow_%d = %d\n", j, *(RGBYchains + i*(t_FVF/I_V) + j)));
		}
  		else if((i == BLUE) && (j%2 == 0))
		{
		    DBG(printf("Blue_%d = %d\n", j, *(RGBYchains + i*(t_FVF/I_V) + j)));
		}
		else if((i == GREEN) && (j%2 == 0))
		{
		    DBG(printf("Green_%d = %d\n", j, *(RGBYchains + i*(t_FVF/I_V) + j)));
		}
            }
        }
    }
    /*end of RGBY data structure initializaton and tampering*/

    /*Initialize and tamper the tiles hash chain data structure for the tiled bitmap algorithm*/
    if(strncmp(alg_flag, "t", FLAG_SIZE) == ZERO)
    {
        tiles = (int *)mymalloc((int)((t_FVF/(I_N))*num_chains_per_tile) * sizeof(int));
        DBG(printf("Initializing the tiles data structure...\n"));
        for( i = 0 ; i < t_FVF/(I_N) ; i++)
        {
            for( j = 0; j < num_chains_per_tile; j++)
            {
	  	counter_tiled_n++;
	        *(tiles + (i*num_chains_per_tile) + j)= 1;    /*initialization of the tiles data structure*/
            }
        }

        unsigned int rel;
        /*Tamper tiles data structure. The validation results 
	should be consistent with the provided CE locus times t_ls*/
        for(i = 0; i < ce_num; i++)
        {
            /*c_0 always evaluates to 0*/
            tiles[(-1+(int)ceil((float)t_l[i] / I_N))*num_chains_per_tile] = 0;
            /*the rest of the chains c_1-c_x must be set*/
            rel = (t_l[i] - 1) % I_N;
            rel = rel/R_s;
            DBG(printf("The relative position within tile of locus time t_l is %d\n", rel));
            for (j = 0; j < num_chains_per_tile-1 ; j++)
            {
                if(!(rel & (1 << j))) /*if the bit is not set yet, then set it*/
                {
                    tiles[(-1+(int)ceil((float)t_l[i] / I_N))*num_chains_per_tile + num_chains_per_tile-(j+1)] = 0;
                }
            }
        }
	/*Print the tiles data structure after tampering has occurred*/
        DBG(printf("The tiles data structure after tampering \ntiles = {\n"));
        for( i = 0 ; i < t_FVF/I_N ; i++)
        {
            DBG(printf("[%d,%d]\n", i*I_N + 1, (i+1)*I_N)); 
            for( j = 0; j < num_chains_per_tile; j++)
            {
                DBG(printf("c_%d = %d \n", j, *(tiles + (i*num_chains_per_tile) + j)));
            }
            DBG(printf("\n"));
        }
        DBG(printf("}\n"));
    }/*end of tiled-bitmap data structure initializaton and tampering*/

    /*Initialize and tamper the tree hash chains data structure for the tiled bitmap algorithm*/
    if(strncmp(alg_flag, "a", FLAG_SIZE) == ZERO)
    {
  	int level, component, m, m_max, level_max, comp_max, t_compl;
    	m_max = 1 << (int)ceil(log_base_n(t_FVF/I_V, 2));  /*last validation of next complete tree*/
    	t_compl = m_max * I_V; /*the time at which the next complete tree will be constructed*/
    	/*level_max: maximum number of levels of hash chain group added at each VE*/
    	level_max = 1 + (int)ceil(log_base_n(N,2)) + (int)floor(log_base_n(m_max,2)); 
    	comp_max = (t_compl/R_s); /*maximum index for the smallest size hash chains at level 0 throughout the entire tree*/

        DBG(printf("The dimensions of the tree hash chains data structure is P[%d][%d][%d]\n", m_max, level_max, comp_max));

        P = (int ***)mymalloc(m_max*sizeof(int***)); /*the 3-D array of the tree hash chains*/

        for(m = 0; m < m_max; m++) 
        {
            P[m] = (int**)mymalloc(sizeof(int**) * level_max);
        }

        for(m = 0; m < m_max; m++) 
            for(level = 0; level < level_max; level++) 
                P[m][level] = (int*)mymalloc(sizeof(int) * comp_max);

        /*initialize the tree hash chains*/
        for(m = 0; m < m_max ; m++)
            for(level = 0; level < level_max ; level++)
            {
                for(component = 0; component < comp_max; component++)
                {
                    P[m][level][component] = -9;
                }
            }

        /*all hash chains belonging to basic tile (i.e., which cover an I_N/I_V) will be initialized to 1*/
        for(m = 0; m < m_max ; m++)
            for(level = 0; level <= (int)log_base_n(N,2) ; level++)
            {
                for(component = (N*m)/(1 << level) ; component < (N*m + N)/(1 << level); component++)
                {
                    if(m >= t_FVF / I_V)
			  P[m][level][component] = NOT_DEFINED;
                    else
                    { 
			  P[m][level][component] = 1;
		    	  counter_a3D_n++;
	  	    }
                }
            }

        DBG(printf("Initializing the completing chains of the tree\n"));
        /*Initialize completing chains*/
        int d;
        component = d = 0;
        for(m = 0; m < m_max ; m++)
        {
            if( ((m+1) % 2) == ZERO)
            {
                d = 0;
                while(((m+1) % (1 << d)) == ZERO)
                    d++;
                d--;
                for( level = 0 ; level <  d ; level++)
                {
                    if(  ( (m+1) % ( 1 << (level+1))) ==  ZERO)
                    {
                        component = (m+1) / ( 1 << (level+1));
                        DBG(printf("Accessing : P[%d,%d,%d]\n",m+1, level+1+(int)log_base_n(N,2), component-1 ));
                        if(!((m+1) > (t_FVF/I_V)))
                        {
			     P[m][level+1+(int)log_base_n(N,2)][component-1] = 1;
  	    		     counter_a3D_n++;
			}
                    }
                }
            }
        }

	/*Tamper the tree data structure. The validation results 
	should be consistent with the provided CE locus times t_ls*/
        d = 0;
        for(j = 0; j < ce_num; j++) /*for each t_l*/
        {
            for(m = 0; m < (t_FVF/I_V) ; m++) /*go up until the last VE -- don't iterate through complete tree*/
            {
                /*tamper within tiles first*/
                for(level = 0; level <= (int)log_base_n(N,2) ; level++)
                {
                    for(component = (N*m)/(1 << level) ; component < (N*m + N)/(1 << level); component++)
                    {
                        if(( ((-1+t_l[j])/R_s) >= ( (1 << level) * component ) )
                            &&( ((-1+t_l[j])/R_s) <= ( (1 << level) *(component+1)-1) ))
                            P[m][level][component] = 0;
                    }
                }
                /* and then tamper completing chains*/
                if( ((m+1) % 2) == ZERO)
                { /* if on an even-numbered VE*/
                    d = component = 0;
                    while(((m+1) % (1 << d)) == ZERO)
                        d++;
                    d--;
                    for( level = 0 ; level < d ; level++)
                    {
                        component = (m+1)*N / (1 << (level+1+(int)log_base_n(N,2))) - 1 ;
                        DBG(printf("Accessing : P[%d,%d,%d]\n",m+1, level+1+(int)log_base_n(N,2), component));
                        if(( ((-1+t_l[j])/R_s) >= ( (1 << (level+1+(int)log_base_n(N,2))) * component ) )
                            && ( ((-1+t_l[j])/R_s) <= ( (1 << (level+1+(int)log_base_n(N,2))) *(component+1)-1) ))
                        {
                            P[m][level+1+(int)log_base_n(N,2)][component] = 0;
                        }
                    }
                }
            }
        }

        /*print tree data structure after tampering*/
        for(m = 0; m < m_max ; m++)
        {
            DBG(printf("\nVE_%d:\n", m+1));
            for(level = 0; level <  level_max ; level++)
            {
                DBG(printf("\tLevel=%d\n", level));
                for(component = 0; component < comp_max; component++)
                {
                    if(P[m][level][component] != NOT_DEFINED)
                        DBG(printf("\t\tP[%d,%d,%d] = %d\n",m+1,level,component, P[m][level][component]));
                }
                DBG(printf("\n"));
            }
            DBG(printf("__________________\n"));
        }
        DBG(printf("\n"));
    }/*end of a3D data structure initializaton and tampering*/

    /*Call the selected algorithm to detect the tampering; report results*/
    if(strncmp(alg_flag,"m",FLAG_SIZE) == ZERO)
    {
	printf("The *Monochromatic* forensic analysis algorithm is called...\n");
        monochromatic(t_FVF, I_N,  V);
	DBG(printf("The number of contacts with the notarization service for the normal processing phase: %d\n", counter_mono_n));
        DBG(printf("The number of contacts with the notarization service for the forensic analysis phase: %d\n", counter_mono_f)); 
    }
    else if(strncmp(alg_flag,"r",FLAG_SIZE) == ZERO)
    {
	printf("The *RGBY* forensic analysis algorithm is called...\n");
	RGBY(t_FVF, I_N, R_s);
	printf("The number of contacts with the notarization service for the normal processing phase: %d\n", counter_RGBY_n);
        printf("The number of contacts with the notarization service for the forensic analysis phase: %d\n", counter_RGBY_f); 

	/*free malloced data structures*/
	free(RGBYchains);	
    }
    else if(strncmp(alg_flag,"t",FLAG_SIZE) == ZERO)
    {
	int default_flag = FALSE;
        printf("Please provide the value of k if known. If not, type '0'.\n");
        scanf("%d", &k);
        if(k == 0)
	{
	    default_flag = TRUE;
            k = 2;
	}
        printf("The *Tiled Bitmap* forensic analysis algorithm is called...\n");
        Tiled_Bitmap(t_FVF, I_N, k, default_flag);
	DBG(printf("The number of contacts with the notarization service for the normal processing phase: %d\n", counter_tiled_n));
        DBG(printf("The number of contacts with the notarization service for the forensic analysis phase: %d\n", counter_tiled_f)); 
	/*free malloced data structures*/
        free(tiles);
        free(cache);
    }
    else
    {
	printf("The *a3D* forensic analysis algorithm is called...\n");
        a3D(t_FVF, I_V, N, R_s);
	DBG(printf("The number of contacts with the notarization service for the normal processing phase: %d\n", counter_a3D_n));
        DBG(printf("The number of contacts with the notarization service for the forensic analysis phase: %d\n", counter_a3D_f)); 

	/*free malloced data structures*/
	int level, m, m_max, level_max;
    	m_max = 1 << (int)ceil(log_base_n(t_FVF/I_V, 2));  /*height of next complete tree*/
        level_max = 1 + (int)ceil(log_base_n(N,2)) + (int)floor(log_base_n(m_max,2)); 
        for(m = 0; m < m_max; m++) 
            for(level = 0; level < level_max; level++)
                free(P[m][level]);
        for(m = 0; m < m_max; m++) 
        {
            free(P[m]);
        }
        free(P);
    }
    free(BlackChains);
    free(t_l);
    return 0;
}

/**************************/
/*	    	          */
/* Helper Functions       */
/*                        */
/**************************/

/*****************************************************************/
/* Input: t_FVF is the time of first validation failure.         */
/*	  I_N is the notarization interval		         */
/* Output: t_RVS is the time of most recent validation success.  */
/* The function performs a binary search on the validation       */
/* results of the black chains and locates the unique transition */
/* from the sequence of TRUEs to the sequence of FALSEs.         */
/* t_RVS is the last validation result which evaluated to TRUE   */
/* (the last in the sequence of TRUEs before the FALSEs start).  */
/*****************************************************************/
int find_t_RVS( int t_FVF, int I_N)
{
    int left, right , t_RVS, BC1, BC2;
    left = 1;
    right = t_FVF;
    t_RVS = (int)floor((left + right)/2.0);
    if(t_RVS % I_N != ZERO)
        t_RVS = t_RVS - (t_RVS % I_N);
    BC1 = BlackChains[MAX(-1+(t_RVS/I_N),0)];
    BC2 = BlackChains[t_RVS/I_N];
    counter_mono_f += 2;
    counter_RGBY_f += 2;
    counter_tiled_f += 2;
    while((!BC1 || BC2) && (right >= left)) /*binary search on black chains*/
    {
        if(!BC2)
        {
            /*corruption on the first NE_0 implies schema corruption*/
            /*if(t_RVS == 0) 
            {
                printf("Schema Corruption: cannot proceed ...\n");
                exit(1);
            }*/
            right = (t_RVS - I_N < 0)? ZERO :t_RVS - I_N;
        }
        else
            left = (t_RVS + I_N > t_FVF) ?  t_FVF : t_RVS + I_N;

        t_RVS = (int)floor((left + right)/2.0);
        if(t_RVS % I_N != ZERO)
            t_RVS = t_RVS - (t_RVS % I_N);      

        BC1 = BlackChains[MAX(-1+(t_RVS/I_N),0)];
        BC2 = BlackChains[t_RVS/I_N];
        counter_mono_f+=2;
	counter_RGBY_f += 2;
        counter_tiled_f+=2;
    }
    return t_RVS;
}

/****************************************************************************/
/* Wrapper function for malloc which performs a check after the system call */
/****************************************************************************/
void * mymalloc(int size)
{
    void *p= malloc(size);
    if(p == NULL)
    {
	fprintf(stderr, "Memory allocation failed\n");
        exit(1);
    }
    bzero(p, size);
    return p;
}

/*************************************************************/
/* Input: num is the number whose logarithm is calculated    */
/*	  base is the base of the logarithm                  */
/* Output: the logarithm of base 'base' of number 'num'      */
/*************************************************************/
float log_base_n(float num, float base)
{
    float  res;
    if(num < 1 )
    {
        res = 0;
    }
    else 
    {
        res = log(num)/log(base);
    }
    return res;
}


